<?php

/**
 * DeubertIT Healthchecks
 *
 * LICENSE: Redistribution and use in source and binary forms, with or without
 * modifications, are not permitted.
 *
 * @copyright Deubert IT-Solutions© 2019 All rights reserved
 */

namespace DeubertIt\Healthchecks;

use Composer\Composer;
use Composer\Script\Event;
use Composer\EventDispatcher\EventSubscriberInterface;
use Composer\IO\IOInterface;
use Composer\Plugin\PluginInterface;
use Composer\Installer\PackageEvent;
use Composer\Package\PackageInterface;

class Healthchecks implements PluginInterface, EventSubscriberInterface
{
    const PACKAGE_NAME = 'deubertit/composer-healthchecks';

    const EXTRA_PUBLIC_PATH = 'dit-healthchecks-public-path';

    const HEALTHCHECKS_SRC_FILENAME = 'healthcheck.dist.php';
    const HEALTHCHECKS_DEST_FILENAME = 'healthcheck.php';
    const HEALTHCHECK_PLACEHOLDER_AUTOLOAD = '##healthchecks-autoload-path##';
    const HEALTHCHECK_PLACEHOLDER_COMPOSER_DIR = '##composer-dir##';

    /**
     * @var Composer
     */
    protected $_composer;

    /**
     * @var IOInterface
     */
    protected $_io;

    /**
     * @var string
     */
    protected $_publicPath;

    /**
     * @return  array       Callbacks list for composer events
     */
    public static function getSubscribedEvents()
    {
        return [
            'post-package-install' => [
                ['onPackageAction', 0]
            ],
            'post-package-update' => [
                ['onPackageAction', 0]
            ]
        ];
    }

    /**
     * @param   Composer            $composer
     * @param   IOInterface         $io
     */
    public function activate(Composer $composer, IOInterface $io)
    {
        $this->_composer = $composer;
        $this->_io = $io;

        $extra = $this->_composer->getPackage()->getExtra();

        if (!isset($extra[self::EXTRA_PUBLIC_PATH]) 
            || empty($extra[self::EXTRA_PUBLIC_PATH]))
        {
            $this->_io->writeError(sprintf(
                '  --- %s: extra.%s value is mandatory!', 
                self::PACKAGE_NAME,
                self::EXTRA_PUBLIC_PATH
            ));

            return;
        }

        $this->_publicPath = rtrim(getcwd() . '/' . $extra[self::EXTRA_PUBLIC_PATH], '/');
    }

    /**
     * @param  PackageEvent         $event
     */
    public function onPackageAction(PackageEvent $event)
    {
        if (empty($this->_publicPath))
        {
            // module is not configured properly
            // error message will be shown before in self::activate()
            return;
        }

        if ($event->getOperation()->getJobType() === 'update')
        {
            $package = $event->getOperation()->getTargetPackage();
        }
        else
        {
            $package = $event->getOperation()->getPackage();
        }

        if ($package->getName() === self::PACKAGE_NAME)
        {
            $composerConfig = $this->_composer->getConfig();

            $packageDir = $composerConfig->get('vendor-dir') . '/' . self::PACKAGE_NAME;
            $srcHealthchecksFile = $packageDir . '/' . self::HEALTHCHECKS_SRC_FILENAME;

            if (!is_file($srcHealthchecksFile) || !($fileContent = file_get_contents($srcHealthchecksFile)))
            {
                $this->_io->writeError(sprintf(
                    '  --- %s: source healthcheck file could not be found in %s!',
                    self::PACKAGE_NAME,
                    $srcHealthchecksFile
                ));

                return;
            }

            $destHealthchecksFile = $this->_publicPath . '/' . self::HEALTHCHECKS_DEST_FILENAME;

            $fileContent = str_replace([
                self::HEALTHCHECK_PLACEHOLDER_AUTOLOAD,
                self::HEALTHCHECK_PLACEHOLDER_COMPOSER_DIR
            ], [
                $this->_getAutoloadRelativePath($destHealthchecksFile, $packageDir),
                $this->_getComposeDirRelativePath($destHealthchecksFile, $composerConfig->get('vendor-dir') . '/../'),
            ], $fileContent);

            if (is_file($destHealthchecksFile) && !unlink($destHealthchecksFile))
            {
                $this->_io->writeError(sprintf(
                    '  --- %s: destination healthcheck file is already exists %s and can not be deleted',
                    self::PACKAGE_NAME,
                    $destHealthchecksFile
                ));

                return;
            }

            if (!is_dir($this->_publicPath))
            {
                mkdir($this->_publicPath, 0777, true);
            }

            if ( ! @file_put_contents($destHealthchecksFile, $fileContent))
            {
                $this->_io->writeError(sprintf(
                    '  --- %s: destination healthcheck file could not be created in %s!',
                    self::PACKAGE_NAME,
                    $destHealthchecksFile
                ));

                return;
            }

            $this->_io->write('  --- healthckecks script successfully installed to ' . $destHealthchecksFile);
        }
    }

    protected function _getAutoloadRelativePath($destHealthchecksFile, $packageDir)
    {
        return '/' . $this->_getRelativePath($destHealthchecksFile, $packageDir . '/autoload.php');
    }

    protected function _getComposeDirRelativePath($destHealthchecksFile, $packageDir)
    {
        return '/' . $this->_getRelativePath($destHealthchecksFile, $packageDir);
    }

    /**
     * @see https://stackoverflow.com/questions/2637945/getting-relative-path-from-absolute-path-in-php
     * 
     * @param   string          $from
     * @param   string          $to
     */
    protected function _getRelativePath($from, $to)
    {
        // some compatibility fixes for Windows paths
        $from = is_dir($from) ? rtrim($from, '\/') . '/' : $from;
        $to   = is_dir($to)   ? rtrim($to, '\/') . '/'   : $to;
        $from = str_replace('\\', '/', $from);
        $to   = str_replace('\\', '/', $to);

        $from     = explode('/', $from);
        $to       = explode('/', $to);
        $relPath  = $to;

        foreach($from as $depth => $dir)
        {
            // find first non-matching dir
            if($dir === $to[$depth])
            {
                // ignore this directory
                array_shift($relPath);
            } 
            else 
            {
                // get number of remaining dirs to $from
                $remaining = count($from) - $depth;
                if($remaining > 1) {
                    // add traversals up to first matching dir
                    $padLength = (count($relPath) + $remaining - 1) * -1;
                    $relPath = array_pad($relPath, $padLength, '..');
                    break;
                } else {
                    $relPath[0] = './' . $relPath[0];
                }
            }
        }

        return implode('/', $relPath);
    }
}
